/**
  *  \file utf8.cc
  *  \brief UTF-8 Encoding and Decoding
  *
  *  A "rune" is an UTF-8 encoded character, i.e. possibly multiple
  *  bytes.
  *
  *  \todo Right now, this supports the BMP only.
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "cpluslib/utf8.h"


/** Get n'th byte from the future. */
inline uint8_t
Utf8Reader::peek(size_t n) const
{
    return text[n];
}

/** Advance by n bytes. */
inline void
Utf8Reader::advance(size_t n)
{
    text += n;
    size -= n;
}

/** Count number of runes in a string. */
size_t
Utf8Reader::count()
{
    Utf8Reader tmp(*this);
    size_t n = 0;
    while (tmp.hasMore()) {
        ++n;
        tmp.get();
    }
    return n;
}

/** Read a rune. Returns the character. If the UTF-8 stream is
    invalid, returns an error character.
    \pre hasMore() */
uint32_t
Utf8Reader::get()
{
    // 0xxxxxxx                   = 0000 .. 007F
    // 110zzzzX 10XXXXXX          = 0080 .. 07FF
    // 1110zzzz 10zXXXXX 10XXXXXX = 0800 .. FFFF
    uint32_t uch;
    uint8_t ch = peek(0);
    if (ch < 0x80) {
        // single byte
        advance(1);
        return ch;
    } else if (ch >= 0xC0 && ch < 0xE0) {
        // two bytes
        if (size >= 2
            && (peek(1) & 0xC0) == 0x80
            && (uch = ((ch & 0x1F) << 6) + (peek(1) & 0x3F)) >= 0x80
            && uch <= 0x7FF)
        {
            // Valid two-byte encoding
            advance(2);
            return uch;
        } else {
            // Invalid two-byte encoding, return error character
            advance(1);
            return makeErrorCharacter(ch);
        }
    } else if (ch >= 0xE0 && ch < 0xF0) {
        // three bytes
        if (size >= 3
            && (peek(1) & 0xC0) == 0x80
            && (peek(2) & 0xC0) == 0x80
            && (uch = ((ch & 0x0F) << 12) + ((peek(1) & 0x3F) << 6) + (peek(2) & 0x3F)) >= 0x800
            && uch <= 0xFFFF)
        {
            // Valid three-byte encoding
            advance(3);
            return uch;
        } else {
            // Invalid three-byte encoding, return error rune
            advance(1);
            return makeErrorCharacter(ch);
        }
    } else {
        // invalid
        advance(1);
        return makeErrorCharacter(ch);
    }
}


/** Append character, encoding as UTF-8.
    \param s  [out] Receives the UTF-8 runes
    \param ch [in] Character to process
    \return s */
string_t&
appendUtf8(string_t& s, uint32_t ch)
{
    if (ch < 0x80) {
        // one byte
        s += char(ch);
    } else if (ch < 0x800) {
        // two bytes
        s += char(0xC0 + (ch >> 6));
        s += char(0x80 + (ch & 0x3F));
    } else if (ch < 0x10000) {
        // three bytes
        s += char(0xE0 + (ch >> 12));
        s += char(0x80 + ((ch >> 6) & 0x3F));
        s += char(0x80 + (ch & 0x3F));
    } else {
        // now what?
    }
    return s;
}

/** Get length of UTF-8 encoded string, in characters.
    \param s the string */
size_t
lengthUtf8(const string_t& s)
{
    return lengthUtf8(s.data(), s.size());
}

/** Get length of UTF-8 encoded string, in characters.
    \param text pointer to text
    \param size number of bytes */
size_t
lengthUtf8(const char* text, size_t size)
{
    return Utf8Reader(text, size).count();
}

/** Get substring of UTF-8 encoded string.
    \param s string
    \param index start extracting at this character
    \param count extract this many characters */
string_t
substrUtf8(const string_t& s, size_t index, size_t count)
{
    return substrUtf8(s.data(), s.size(), index, count);
}

/** Get substring of UTF-8 encoded string.
    \param text pointer to text
    \param size number of bytes
    \param index start extracting at this character
    \param count extract this many characters */
string_t
substrUtf8(const char* text, size_t size, size_t index, size_t count)
{
    // Skip over initial portion
    Utf8Reader r(text, size);
    while (r.hasMore() && index != 0) {
        r.get();
        --index;
    }

    if (count >= r.getRemainingSize()) {
        // User wants the whole remaining string
        return string_t(r.getRemainingData(), r.getRemainingSize());
    } else {
        // User wants possibly less than the whole string
        const char* p = r.getRemainingData();
        while (r.hasMore() && count != 0) {
            r.get();
            --count;
        }
        return string_t(p, r.getRemainingData() - p);
    }
}

/** Get character from UTF-8 encoded string.
    \param s String
    \param index 0-based index
    \return Character; 0 if index is out-of-bounds */
uint32_t
charAtUtf8(const string_t& s, size_t index)
{
    return charAtUtf8(s.data(), s.size(), index);
}

/** Get character from UTF-8 encoded string.
    \param text pointer to text
    \param size number of bytes
    \param index 0-based index
    \return Character; 0 if index is out-of-bounds */
uint32_t
charAtUtf8(const char* text, size_t size, size_t index)
{
    // Skip over initial portion
    Utf8Reader r(text, size);
    while (r.hasMore() && index != 0) {
        r.get();
        --index;
    }

    // Check bounds
    if (index == 0 && r.hasMore())
        return r.get();
    else
        return 0;
}

/** Given a byte position, return the corresponding character position. */
size_t
byteToCharPosUtf8(const string_t& s, size_t bytePos)
{
    Utf8Reader rdr(s.data(), s.size());
    size_t total = s.size();
    size_t result = 0;
    while (total - rdr.getRemainingSize() < bytePos) {
        rdr.get();
        ++result;
    }
    return result;
}

/** Given a character number, return the corresponding byte position. */
size_t
charToBytePosUtf8(const string_t& s, size_t charPos)
{
    if (charPos >= s.size()) {
        return s.size();
    } else {
        Utf8Reader rdr(s.data(), s.size());
        size_t total = s.size();
        while (charPos != 0 && rdr.getRemainingSize() != 0) {
            rdr.get();
            --charPos;
        }
        return total - rdr.getRemainingSize();
    }
}

/** Convert uint16_t array to UTF-8.
    \param text Characters
    \param count Number of characters
    \return UTF-8 string */
string_t
convertU16ToUtf8(const uint16_t* text, size_t count)
{
    string_t result;
    while (count--)
        appendUtf8(result, *text++);
    return result;
}

/** Convert wchar_t array to UTF-8.
    \param text Characters
    \param count Number of characters
    \return UTF-8 string */
string_t
convertWcharToUtf8(const wchar_t* text, size_t count)
{
    string_t result;
    while (count--)
        appendUtf8(result, *text++);
    return result;
}

/** Convert Latin-1 text to UTF-8.
    \param text Characters
    \param count Number of characters
    \return UTF-8 string */
string_t
convertLatin1toUtf8(const char* text, size_t count)
{
    string_t result;
    while (count--)
        appendUtf8(result, uint8_t(*text++));
    return result;
}

/** Convert codepage text to UTF-8.
    \param text Characters
    \param count Number of characters
    \param cp Codepage to use (array mapping the upper 128 bytes to Unicode)
    \return UTF-8 string */
string_t
convertCodepageToUtf8(const char* text, size_t count, const Codepage& cp)
{
    string_t result;
    while (count--) {
        uint8_t ch = *text++;
        if (ch >= 0x80)
            appendUtf8(result, cp[ch-0x80]);
        else
            appendUtf8(result, ch);
    }
    return result;
}
